from __future__ import annotations

import platform

import pythoncom
import wmi
from PySide2 import QtWidgets, QtGui, QtCore

from hdlg.hdd import HDD
from hdlg.ui import BaseWindow
from hdlg.utils import size_unit


class Main(BaseWindow):
    def __init__(self):
        super().__init__(name=self.__class__.__name__)

        self.window.setMinimumSize(1000, 400)
        self.window.installButton.setEnabled(False)
        self.window.installButton.hide()
        self.window.progressBar.hide()
        self.window.hddInfoList.header().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)

        self.window.actionExit.triggered.connect(self.window.close)
        self.window.actionAbout.triggered.connect(self.about)
        self.window.refreshIcon.clicked.connect(self.get_hdd_list)

        self.get_hdd_list()

    def clear_hdd_list(self) -> None:
        """Clear all buttons from the HDD list."""
        for child in self.window.deviceListDevices_2.children():
            if isinstance(child, QtWidgets.QPushButton):
                # noinspection PyTypeChecker
                child.setParent(None)

    def add_hdd_button(self, hdd: HDD) -> None:
        """Add a HDD button into the HDD list. Skips buttons with identical targets."""
        for child in self.window.deviceListDevices_2.children():
            if isinstance(child, QtWidgets.QPushButton):
                if child.objectName() == hdd.target:
                    return

        button = QtWidgets.QPushButton("\n".join([
            " ".join([
                size_unit(hdd.disk_size) + ":",
                hdd.target.upper().replace(r"\\.\PHYSICALDRIVE", "HDD"),
                ["", "(PS2)"][hdd.is_apa_partitioned]
            ]),
            hdd.model
        ]))
        button.setObjectName(hdd.target)
        button.setCursor(QtGui.QCursor(QtCore.Qt.PointingHandCursor))
        button.setEnabled(hdd.is_apa_partitioned)
        button.clicked.connect(lambda: self.load_hdd(hdd))

        device_list = self.window.deviceListDevices_2.layout()
        device_list.insertWidget(0 if hdd.is_apa_partitioned else device_list.count() - 1, button)

    def get_hdd_list(self) -> None:
        """Finds HDD devices and adds them to the HDD list."""
        self.thread = QtCore.QThread()
        self.worker = MainWorker()
        self.worker.moveToThread(self.thread)

        self.worker.finished.connect(self.thread.quit)
        self.worker.finished.connect(self.worker.deleteLater)
        self.thread.finished.connect(self.thread.deleteLater)

        def manage_state():
            self.window.refreshIcon.setEnabled(False)
            self.window.installButton.hide()
            self.window.hddInfoList.clear()
            self.window.hddInfoList.setEnabled(False)
            self.clear_hdd_list()
            self.window.statusbar.showMessage("Scanning HDDs...")

            self.window.hddInfoList.addTopLevelItem(QtWidgets.QTreeWidgetItem([
                "\n" * 8 + " " * 90 +
                "Scanning for PS2 HDDs..."
            ]))

        def on_finish(n: int):
            self.window.refreshIcon.setEnabled(True)
            self.window.hddInfoList.clear()
            self.window.statusbar.showMessage("Found %d HDDs" % n)

            self.window.hddInfoList.addTopLevelItem(QtWidgets.QTreeWidgetItem([
                "\n" * 8 + " " * 60 +
                "Ready to go? Just choose a PS2 HDD to get started!"
            ]))

        def on_error(e: Exception):
            print("An error occurred somewhere in Main->get_hdd_list():", e)

        self.thread.started.connect(manage_state)
        self.worker.finished.connect(on_finish)
        self.worker.error.connect(on_error)
        self.worker.found_device.connect(self.add_hdd_button)

        self.thread.started.connect(self.worker.find_hdds)
        self.thread.start()

    def load_hdd(self, hdd: HDD):
        """Load HDD device, get HDD object, get device information."""
        self.thread = QtCore.QThread()
        self.worker = MainWorker()
        self.worker.moveToThread(self.thread)

        self.worker.finished.connect(self.thread.quit)
        self.worker.finished.connect(self.worker.deleteLater)
        self.thread.finished.connect(self.thread.deleteLater)

        def manage_state():
            self.window.deviceListDevices_2.setEnabled(False)
            self.window.refreshIcon.setEnabled(False)
            self.window.installButton.hide()
            self.window.hddInfoList.clear()
            self.window.statusbar.showMessage(f"Loading HDD %s (%s)" % (hdd.target, hdd.model))

            if self.window.installButton.isEnabled():
                self.window.installButton.clicked.disconnect()

        def on_finish(_: int):
            self.window.deviceListDevices_2.setEnabled(True)
            self.window.refreshIcon.setEnabled(True)
            self.window.installButton.setEnabled(True)
            self.window.installButton.show()
            self.window.hddInfoList.setEnabled(True)
            self.window.statusbar.showMessage("Loaded HDD %s (%s)" % (hdd.target, hdd.model))

        def on_error(e: Exception):
            print("An error occurred somewhere in Main->load_device():", e)

        def use_hdd_info(trees: list[QtWidgets.QTreeWidgetItem]):
            self.window.hddInfoList.clear()
            for tree in trees:
                self.window.hddInfoList.addTopLevelItem(tree)

            self.window.hddInfoList.expandToDepth(0)

        self.thread.started.connect(manage_state)
        self.worker.finished.connect(on_finish)
        self.worker.error.connect(on_error)
        self.worker.hdd_info.connect(use_hdd_info)

        self.thread.started.connect(lambda: self.worker.get_hdd_info(hdd))
        self.thread.start()


class MainWorker(QtCore.QObject):
    error = QtCore.Signal(Exception)
    finished = QtCore.Signal(int)
    found_device = QtCore.Signal(HDD)
    hdd_info = QtCore.Signal(list)

    def find_hdds(self) -> None:
        """Find Disk Drive devices using win32 api on Windows, or lsscsi on Linux."""
        try:
            # TODO: Support Linux and macOS
            system = platform.system()
            if system == "Windows":
                # noinspection PyUnresolvedReferences
                pythoncom.CoInitialize()  # important!
                c = wmi.WMI()
                disk_drives = c.Win32_DiskDrive()
                for disk_drive in sorted(disk_drives, key=lambda d: d.index, reverse=True):
                    self.found_device.emit(HDD(
                        target=disk_drive.DeviceID,
                        model=disk_drive.Model
                    ))
                self.finished.emit(len(disk_drives))
            else:
                raise NotImplementedError("Device Scanning has not been implemented for %s." % system)
        except Exception as e:
            self.error.emit(e)

    def get_hdd_info(self, hdd: HDD) -> None:
        try:
            games = hdd.get_games_list()

            games_tree = QtWidgets.QTreeWidgetItem(["Games", str(len(games))])
            for media_type, size, _, dma, game_id, name in games:
                games_tree.addChild(QtWidgets.QTreeWidgetItem([
                    f"{media_type} {size_unit(size)} ({dma})",
                    f"{game_id} {name}"
                ]))

            self.hdd_info.emit([
                QtWidgets.QTreeWidgetItem(["Disk Size", f"{size_unit(hdd.disk_size)} ({hdd.disk_size})"]),
                games_tree
            ])
            self.finished.emit(0)
        except Exception as e:
            self.error.emit(e)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hdlg', 'hdlg.ui']

package_data = \
{'': ['*'], 'hdlg.ui': ['icons/*']}

install_requires = \
['PySide2>=5.15.2,<6.0.0', 'appdirs>=1.4.4,<2.0.0', 'toml>=0.10.2,<0.11.0']

extras_require = \
{':sys_platform == "win32"': ['pywin32==302', 'WMI>=1.5.1,<2.0.0']}

entry_points = \
{'console_scripts': ['hdlg = hdlg.hdlg:main']}

setup_kwargs = {
    'name': 'hdlg',
    'version': '0.1.0',
    'description': 'Modern cross-platform GUI for hdl-dump.',
    'long_description': "# HDLG\n\n[![GPLv3 license](https://img.shields.io/badge/license-GPLv3-blue)](https://github.com/rlaphoenix/hdlg/blob/master/LICENSE)\n[![PyPI version](https://img.shields.io/pypi/v/hdlg)](https://pypi.org/project/hdlg)\n[![Python versions](https://img.shields.io/pypi/pyversions/hdlg)](https://pypi.org/project/hdlg)\n[![PyPI status](https://img.shields.io/pypi/status/hdlg)](https://pypi.org/project/hdlg)\n[![Contributors](https://img.shields.io/github/contributors/rlaphoenix/hdlg)](https://github.com/rlaphoenix/hdlg/graphs/contributors)\n[![GitHub issues](https://img.shields.io/github/issues/rlaphoenix/hdlg)](https://github.com/rlaphoenix/hdlg/issues)\n![Build](https://github.com/rlaphoenix/hdlg/workflows/Build/badge.svg?branch=master)\n\nHDLG is a modern cross-platform GUI for hdl-dump with Batch installation capabilities.\n\n## Looking for Artwork\n\nThis project is looking for an Icon and Text Logo as well as a Banner artwork. If you have some free time and would\nlike to contribute artwork to the project, let me know!\n\n## Installation\n\n    pip install --user hdlg\n\nTo run hdlg, type `hdlg` into any terminal, command prompt, app launcher, or the start menu.\n\nIf you wish to manually install from the source, take a look at [Building](#building-source-and-wheel-distributions).\n\n## To-do\n\n- [x] Craft initial GUI with Qt.\n- [x] Push to PyPI and add relevant Badges.\n- [x] Add PyInstaller make file.\n- [x] Add local PS2 HDD connection option.\n- [x] List installed games of selected HDD.\n- [ ] Show HDD information like Disk Size, Space Used, and such.\n- [ ] Add ability to install a new game to selected HDD.\n- [ ] Create a file based settings system.\n- [ ] Add remote PS2 HDD (samba) connection option.\n\n## Building\n\nThis project requires [Poetry], so feel free to take advantage and use it for its various conveniences like\nbuilding sdist/wheel packages, creating and managing dependencies, virtual environments, and more.\n\nNote:\n\n- Source Code may have changes that may be old, not yet tested or stable, or may have regressions.\n- Only run or install from Source Code if you have a good reason. Examples would be to test for regressions, test\n  changes (either your own or other contributors), or to research the code (agreeing to the [LICENSE](LICENSE)).\n- [Poetry] is required as it's used as the [PEP 517] build system, virtual environment manager, dependency manager,\n  and more.\n\n  [Poetry]: <https://python-poetry.org/docs/#installation>\n  [PEP 517]: <https://www.python.org/dev/peps/pep-0517>\n\n### Install from Source Code\n\n    git clone https://github.com/rlaphoenix/hdlg.git\n    cd hdlg\n    pip install --user .\n\n### Building source and wheel distributions\n\n    poetry build\n\nYou can specify `-f` to build `sdist` or `wheel` only. Built files can be found in the `/dist` directory.\n\n### Packing with PyInstaller\n\n    poetry run python pyinstaller.py\n\nThe build is now available at `./dist`.\n",
    'author': 'PHOENiX',
    'author_email': 'rlaphoenix@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rlaphoenix/hdlg',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<3.10',
}


setup(**setup_kwargs)



"""
hdlg - Modern cross-platform GUI for hdl-dump.
Copyright (C) 2021 rlaphoenix

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

from __future__ import annotations

import struct
import subprocess
import sys
from ctypes.wintypes import HANDLE
from pathlib import Path
from typing import Union

import win32con
import win32file
import winioctlcon

from hdlg.utils import NEIGHBORING_WHITESPACE


class HDD:
    def __init__(self, target: Union[str, Path], model: str):
        self.handle = win32file.INVALID_HANDLE_VALUE
        self.target = target
        self.model = model

        self._geometry = None
        self._disk_size = None
        self._is_apa_partitioned = None
        self._apa_checksum = None

        self.open(target)

    def __enter__(self):
        return self

    def __exit__(self, *_, **__):
        self.dispose()

    def dispose(self):
        if self.handle != win32file.INVALID_HANDLE_VALUE:
            win32file.CloseHandle(self.handle)

    def open(self, device: str, extended=False) -> HANDLE:
        if sys.platform == "win32":
            if not device.startswith("\\\\.\\"):
                device = r"\\.\%s" % device  # unc target
        self.handle = win32file.CreateFile(
            # https://docs.microsoft.com/en-us/windows/win32/api/fileapi/nf-fileapi-createfilea
            device,  # target
            win32con.MAXIMUM_ALLOWED,  # 0x0080 | 0x0020,  # desired access
            win32con.FILE_SHARE_READ | win32con.FILE_SHARE_WRITE,  # share mode
            None,  # security attributes
            win32con.OPEN_EXISTING,  # creation disposition
            win32con.FILE_ATTRIBUTE_NORMAL,  # flags and attributes
            None  # template file
        )
        if self.handle == win32file.INVALID_HANDLE_VALUE:
            raise RuntimeError("Failed to obtain device handle...")
        if extended:
            # TODO: Sometimes works, sometimes doesn't. Only got it working (once) on a CD-ROM drive with a DVD.
            print(win32file.DeviceIoControl(self.handle, winioctlcon.FSCTL_ALLOW_EXTENDED_DASD_IO, None, None))
        return self.handle

    def seek(self, to: int, whence: int = win32file.FILE_BEGIN) -> int:
        pos = win32file.SetFilePointer(self.handle, to, whence)
        if pos != to:
            raise IOError(f"Seek was not precise...")
        return pos

    def read(self, size: int) -> bytes:
        if size % 512 != 0:
            raise ValueError("Size must be a multiple of 512 for some reason, Ask Windows.")
        res, data = win32file.ReadFile(self.handle, size, None)
        if res != 0:
            raise IOError(f"An error occurred: {res} {data}")
        if len(data) < size:
            raise IOError(f"Read {size - len(data)} less bytes than requested...")
        return data

    @property
    def geometry(self) -> tuple[int, ...]:
        """
        Retrieves information about the physical disk's geometry.
        https://docs.microsoft.com/en-us/windows/win32/api/winioctl/ns-winioctl-disk_geometry_ex

        Returns a tuple of:
            Cylinders-Lo
            Cylinders-Hi
            Media Type
            Tracks Per Cylinder
            Sectors Per Track
            Bytes Per Sector
            Disk Size
            Extra Data
        """
        if self._geometry is not None:
            return self._geometry

        if self.handle == win32file.INVALID_HANDLE_VALUE:
            raise ValueError("Handle is invalid, have you opened a handle yet?")

        self._geometry = struct.unpack("8L", win32file.DeviceIoControl(
            self.handle,  # handle
            winioctlcon.IOCTL_DISK_GET_DRIVE_GEOMETRY_EX,  # ioctl api
            None,  # in buffer
            32  # out buffer
        ))
        return self._geometry

    @property
    def disk_size(self) -> int:
        """Get full Disk Size (in bytes)."""
        if self._disk_size is not None:
            return self._disk_size

        cyl_lo, cyl_hi, _, tpc, spt, bps, _, _ = self.geometry

        self._disk_size = (cyl_lo + cyl_hi) * tpc * spt * bps
        return self._disk_size

    @property
    def is_apa_partitioned(self) -> bool:
        """Check if HDD is a PS2 APA-formatted device."""
        if self._is_apa_partitioned is not None:
            return self._is_apa_partitioned

        old_pos = self.seek(0, whence=win32file.FILE_CURRENT)

        try:
            if self.seek(0) != 0:
                raise RuntimeError("Unable to seek to start of HDD, cannot check if APA Partitioned...")
            header = self.read(1024)
            checksum = header[0:4]
            magic = header[4:8]

            new_checksum = struct.pack(
                "<Q",
                sum(struct.unpack("<I", header[n:n+4])[0] for n in range(4, 1024, 4))
            )[:4]

            self._is_apa_partitioned = magic == b"APA\0" and checksum == new_checksum
        finally:
            self.seek(old_pos)

        return self._is_apa_partitioned

    @property
    def apa_checksum(self):
        if not self.is_apa_partitioned:
            return None

        if self._apa_checksum is not None:
            return self._apa_checksum

        old_pos = self.seek(0, whence=win32file.FILE_CURRENT)

        try:
            if self.seek(0) != 0:
                raise RuntimeError("Unable to seek to start of HDD, cannot get APA checksum...")
            header = self.read(512)
            self._apa_checksum = header[0:4]
        finally:
            self.seek(old_pos)

        return self._apa_checksum

    def get_games_list(self) -> list[tuple[str, int, int, str, str, str]]:
        """
        Get a list of games installed on the HDD (if any).

        Returns a tuple of:
            MediaType
            Size (in Bytes)
            Flags
            DMA
            GameID
            GameName
        """
        # TODO: Add handler for when there's no games installed
        hdl_toc = subprocess.check_output([
            "hdl_dump.exe",
            "hdl_toc",
            self.target.upper().replace(r"\\.\PHYSICALDRIVE", "hdd") + ":"
        ])
        games = hdl_toc.decode().splitlines(keepends=False)[1:-1]
        games = [
            NEIGHBORING_WHITESPACE.sub(" ", game).split(" ", maxsplit=5)
            for game in games
        ]
        games = [
            (media_type, int(size.replace("KB", "")) * 1000, int(flags), dma, game_id, game_name)
            for media_type, size, flags, dma, game_id, game_name in games
        ]
        return games
